typedef union
{
    int num;
    char *ptr;
} YYSTYPE;
#define	LB	258
#define	RB	259
#define	LP	260
#define	RP	261
#define	MENUS	262
#define	MENU	263
#define	BUTTON	264
#define	DEFAULT_FUNCTION	265
#define	PLUS	266
#define	MINUS	267
#define	ALL	268
#define	OR	269
#define	CURSORS	270
#define	PIXMAPS	271
#define	ICONS	272
#define	COLOR	273
#define	SAVECOLOR	274
#define	MONOCHROME	275
#define	FUNCTION	276
#define	ICONMGR_SHOW	277
#define	ICONMGR	278
#define	WINDOW_FUNCTION	279
#define	ZOOM	280
#define	ICONMGRS	281
#define	ICONMGR_GEOMETRY	282
#define	ICONMGR_NOSHOW	283
#define	MAKE_TITLE	284
#define	ICONIFY_BY_UNMAPPING	285
#define	DONT_ICONIFY_BY_UNMAPPING	286
#define	NO_TITLE	287
#define	AUTO_RAISE	288
#define	NO_HILITE	289
#define	ICON_REGION	290
#define	META	291
#define	SHIFT	292
#define	LOCK	293
#define	CONTROL	294
#define	WINDOW	295
#define	TITLE	296
#define	ICON	297
#define	ROOT	298
#define	FRAME	299
#define	VIRTUAL	300
#define	VIRTUAL_WIN	301
#define	COLON	302
#define	EQUALS	303
#define	SQUEEZE_TITLE	304
#define	DONT_SQUEEZE_TITLE	305
#define	START_ICONIFIED	306
#define	NO_TITLE_HILITE	307
#define	TITLE_HILITE	308
#define	MOVE	309
#define	RESIZE	310
#define	WAIT	311
#define	SELECT	312
#define	KILL	313
#define	LEFT_TITLEBUTTON	314
#define	RIGHT_TITLEBUTTON	315
#define	NUMBER	316
#define	KEYWORD	317
#define	NKEYWORD	318
#define	CKEYWORD	319
#define	CLKEYWORD	320
#define	FKEYWORD	321
#define	FSKEYWORD	322
#define	SKEYWORD	323
#define	DKEYWORD	324
#define	JKEYWORD	325
#define	WINDOW_RING	326
#define	WARP_CURSOR	327
#define	ERRORTOKEN	328
#define	NO_STACKMODE	329
#define	NAILEDDOWN	330
#define	VIRTUALDESKTOP	331
#define	NO_SHOW_IN_DISPLAY	332
#define	DOORS	333
#define	DOOR	334
#define	STRING	335


extern YYSTYPE yylval;
